# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/control/control.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5ostorlab/agent/message/proto/v3/control/control.proto\x12\'ostorlab.agent.message.proto.v3.control\"\x19\n\x07\x43ontrol\x12\x0e\n\x06\x61gents\x18\x01 \x03(\t\"]\n\x07Message\x12\x41\n\x07\x63ontrol\x18\x01 \x01(\x0b\x32\x30.ostorlab.agent.message.proto.v3.control.Control\x12\x0f\n\x07message\x18\x02 \x01(\x0c')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.control.control_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CONTROL._serialized_start=98
  _CONTROL._serialized_end=123
  _MESSAGE._serialized_start=125
  _MESSAGE._serialized_end=218
# @@protoc_insertion_point(module_scope)



"""Runtime agent module gives access to methods to create the specific agent configurations & run the agent.

Usage
    agent_runtime = AgentRuntime(agent_settings, runtime_name, docker_client, mq_service, redis_service, jaeger_service)
    agent_service = agent_runtime.create_agent_service(network_name, extra_configs)
"""

import base64
import hashlib
import io
import logging
import random
import uuid
from typing import List, Optional

import docker
from docker import constants
from docker import errors
from docker.types import services as docker_types_services

from ostorlab import configuration_manager
from ostorlab import exceptions
from ostorlab.agent import definitions as agent_definitions
from ostorlab.runtimes import definitions
from ostorlab.runtimes.local.services import jaeger
from ostorlab.runtimes.local.services import mq
from ostorlab.runtimes.local.services import redis

logger = logging.getLogger(__name__)

MOUNT_VARIABLES = {"$CONFIG_HOME": str(configuration_manager.OSTORLAB_PRIVATE_DIR)}

MAX_SERVICE_NAME_LEN = 63
MAX_RANDOM_NAME_LEN = 5


class Error(exceptions.OstorlabError):
    """Base Error."""


class MissingAgentDefinitionLabel(Error):
    """Agent definition label is missing from the agent image. This is likely due to the agent built directly with a
    docker command and not agent build."""


def _parse_mount_string_windows(string):
    """Handles parsing mounts on Windows OS."""
    parts = string.split(":")
    if len(parts) == 1:
        # This supposes only having a target like /root.
        return docker_types_services.Mount(target=parts[0], source=None)
    elif len(parts) == 2:
        # This supposes having a source linux style and target. Like /var/run/docker.sock:/var/run/docker.sock.
        target = parts[1]
        source = parts[0]
        mount_type = "volume"
        if source.startswith("/"):
            # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
            source = f"/{source}"
            mount_type = "bind"
        return docker_types_services.Mount(
            target, source, read_only=False, type=mount_type
        )
    elif len(parts) == 3:
        # This supposes two cases. First case is windows style with linux, like C:/Users/bob:/root. The second
        # case is /root:/root:ro
        if parts[2] in ("ro", "rw", "z", "Z"):
            target = parts[1]
            source = parts[0]
            mount_type = "volume"
            if source.startswith("/"):
                # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
                source = f"/{source}"
                mount_type = "bind"
            read_only = not parts[2] == "rw"
            return docker_types_services.Mount(
                target, source, read_only=read_only, type=mount_type
            )
        else:
            target = parts[2]
            source = ":".join(parts[:2])
            source = source.replace("\\", "/")
            mount_type = "bind"
            return docker_types_services.Mount(
                target, source, read_only=False, type=mount_type
            )
    elif len(parts) == 4:
        # This covers the case C:/Users/bob:/root:ro
        target = parts[2]
        source = ":".join(parts[:2])
        source = source.replace("\\", "/")
        mount_type = "bind"
        read_only = not (len(parts) == 3 or parts[3] == "rw")
        return docker_types_services.Mount(
            target, source, read_only=read_only, type=mount_type
        )
    else:
        raise errors.InvalidArgument(f'Invalid mount format "{string}"')


def _parse_mount_string_unix(string):
    """Handles parsing of mounts on Unix OS like Linux, Cygwin and MacOS."""
    parts = string.split(":")
    if len(parts) > 3:
        raise errors.InvalidArgument(f'Invalid mount format "{string}"')
    if len(parts) == 1:
        return docker_types_services.Mount(target=parts[0], source=None)
    else:
        target = parts[1]
        source = parts[0]
        mount_type = "volume"
        if source.startswith("/"):
            mount_type = "bind"
        read_only = not (len(parts) == 2 or parts[2] == "rw")
        return docker_types_services.Mount(
            target, source, read_only=read_only, type=mount_type
        )


def _parse_mount_string(string):
    """This is a fix to a bug in the Docker Python API by monkey patching the buggy method.

    The method is knowingly not supporting Windows. Until a fix PR is merged, we are hot patching the method to handle
    windows paths with : like C:/Users/bob/:/root.
    """
    if constants.IS_WINDOWS_PLATFORM:
        return _parse_mount_string_windows(string)
    else:
        return _parse_mount_string_unix(string)


docker_types_services.Mount.parse_mount_string = _parse_mount_string


class AgentRuntime:
    """Class to consolidate the agent settings and agent default definition, and create the agent service."""

    def __init__(
        self,
        agent_settings: definitions.AgentSettings,
        runtime_name: str,
        docker_client: docker.DockerClient,
        mq_service: mq.LocalRabbitMQ,
        redis_service: redis.LocalRedis,
        jaeger_service: jaeger.LocalJaeger,
        gcp_logging_credential: Optional[str] = None,
    ) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            agent: an agent definition containing all the settings of how agent should run and what arguments to pass.
            runtime_name: local runtime instance name.
            docker_client: docker client.
        """
        # Unique identifier to allow running multiple instance of the same agent with different configs.
        self._uuid = uuid.uuid4()
        self._docker_client = docker_client
        self.agent = agent_settings
        self.image_name = agent_settings.container_image.split(":", maxsplit=1)[0]
        self.runtime_name = runtime_name
        self.mq_service = mq_service
        self.redis_service = redis_service
        self.jaeger_service = jaeger_service
        self._gcp_logging_credential = gcp_logging_credential
        self.update_agent_settings()

    def create_settings_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent settings.

        Returns:
            docker ConfigReference of the settings configuration
        """
        agent_instance_settings_proto = self.agent.to_raw_proto()
        # The name is hashed to work around size limitation of Docker Config.
        config_name = hashlib.md5(
            f"config_settings_{self.image_name}_{self.runtime_name}_{self._uuid}".encode()
        ).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning(
                "found existing config %s, config will removed", config_name
            )
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug("all good, config %s is new", config_name)

        docker_config = self._docker_client.configs.create(
            name=config_name,
            labels={"ostorlab.universe": self.runtime_name},
            data=agent_instance_settings_proto,
        )
        return docker.types.ConfigReference(
            config_id=docker_config.id,
            config_name=config_name,
            filename="/tmp/settings.binproto",
        )

    def create_definition_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent definition.

        Returns:
            docker configuration reference of the agent defintion configuration.
        """
        agent_definition = self._docker_client.images.get(
            self.agent.container_image
        ).labels.get("agent_definition")
        # The name is hashed to work around size limitation of Docker Config.
        config_name = hashlib.md5(
            f"config_definition_{self.image_name}__{self.runtime_name}_{self._uuid}".encode()
        ).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning(
                "found existing config %s, config will removed", config_name
            )
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug("all good, config %s is new", config_name)

        docker_config = self._docker_client.configs.create(
            name=config_name,
            labels={"ostorlab.universe": self.runtime_name},
            data=str.encode(agent_definition),
        )
        return docker.types.ConfigReference(
            config_id=docker_config.id,
            config_name=config_name,
            filename="/tmp/ostorlab.yaml",
        )

    def create_agent_definition_from_label(self) -> agent_definitions.AgentDefinition:
        """Read the agent yaml definition from the docker image labels.

        Returns:
            the agent definition.
        """
        docker_image = self._docker_client.images.get(self.agent.container_image)
        yaml_definition_string = docker_image.labels.get("agent_definition")
        if yaml_definition_string is None:
            raise MissingAgentDefinitionLabel(
                f"agent definition label is missing from image {docker_image.tags[0]}"
            )
        with io.StringIO(yaml_definition_string) as file:
            agent_definition = agent_definitions.AgentDefinition.from_yaml(file)
            return agent_definition

    def update_agent_settings(self) -> None:
        """Update agent settings with values from the local runtime."""
        self.agent.bus_url = self.mq_service.url
        self.agent.bus_exchange_topic = f"ostorlab_topic_{self.runtime_name}"
        self.agent.bus_management_url = self.mq_service.management_url
        self.agent.bus_vhost = self.mq_service.vhost
        self.agent.redis_url = self.redis_service.url
        self.agent.tracing_collector_url = (
            self.jaeger_service.url if self.jaeger_service is not None else None
        )

    def replace_variable_mounts(self, mounts: List[str]):
        """Replace path variables for the container mounts

        Args:
            mounts: List of src:dst paths to mount
        """

        replaced_mounts = []
        for mount in mounts:
            for mount_variable, mount_value in MOUNT_VARIABLES.items():
                mount = mount.replace(mount_variable, mount_value)
            replaced_mounts.append(mount)
        return replaced_mounts

    def create_agent_service(
        self,
        network_name: str,
        extra_configs: Optional[List[docker.types.ConfigReference]] = None,
        extra_mounts: Optional[List[docker.types.Mount]] = None,
        replicas: int = 1,
    ) -> docker.models.services.Service:
        """Create the docker agent service with proper configs and policies.

        Args:
            replicas:  number of replicas to create
            network_name: network name to attach the service to.
            extra_configs: list of docker ConfigReferences that will be exposed to the service.

        Returns:
            the agent docker service.
        """
        agent_definition = self.create_agent_definition_from_label()
        self.agent.open_ports = self.agent.open_ports or agent_definition.open_ports
        if self.agent.open_ports:
            endpoint_spec = docker_types_services.EndpointSpec(
                ports={p.destination_port: p.source_port for p in self.agent.open_ports}
            )
        else:
            endpoint_spec = docker_types_services.EndpointSpec(mode="dnsrr")

        configs = []
        configs.append(self.create_settings_config())
        configs.append(self.create_definition_config())
        if extra_configs is not None:
            configs.extend(extra_configs)

        mounts = self.agent.mounts or agent_definition.mounts
        mounts = self.replace_variable_mounts(mounts)
        if extra_mounts is not None:
            mounts.extend(extra_mounts)

        constraints = self.agent.constraints or agent_definition.constraints
        mem_limit = self.agent.mem_limit or agent_definition.mem_limit
        restart_policy = (
            self.agent.restart_policy or agent_definition.restart_policy or "any"
        )
        caps = self.agent.caps or agent_definition.caps or []

        service_name = (
            agent_definition.service_name
            or self.agent.container_image.split(":")[0].replace(".", "")
            + "_"
            + self.runtime_name
        )

        # We apply the random str only if it will not break the max docker service name characters (63)
        if len(service_name) + MAX_RANDOM_NAME_LEN < MAX_SERVICE_NAME_LEN:
            service_name = service_name + "_" + str(random.randrange(0, 9999))

        env = [
            f"UNIVERSE={self.runtime_name}",
        ]
        if self._gcp_logging_credential is not None:
            env.append(
                f"GCP_LOGGING_CREDENTIAL={base64.b64encode(self._gcp_logging_credential.encode()).decode()}"
            )

        agent_service = self._docker_client.services.create(
            image=self.agent.container_image,
            networks=[network_name],
            env=env,
            name=service_name,
            restart_policy=docker_types_services.RestartPolicy(
                condition=restart_policy
            ),
            mounts=mounts,
            labels={"ostorlab.universe": self.runtime_name},
            configs=configs,
            constraints=constraints,
            endpoint_spec=endpoint_spec,
            resources=docker_types_services.Resources(mem_limit=mem_limit),
            cap_add=caps,
            mode=docker_types_services.ServiceMode("replicated", replicas),
        )

        return agent_service



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/fingerprint/domain_name/service/x509/x509.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ostorlab.agent.message.proto.common.x509 import x509_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_common_dot_x509_dot_x509__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nOostorlab/agent/message/proto/v3/fingerprint/domain_name/service/x509/x509.proto\x12\x44ostorlab.agent.message.proto.v3.fingerprint.domain_name.service.x509\x1a\x33ostorlab/agent/message/proto/common/x509/x509.proto\"}\n\x07Message\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x0e\n\x06schema\x18\x03 \x01(\t\x12\x46\n\ncert_chain\x18\x04 \x03(\x0b\x32\x32.ostorlab.agent.message.proto.common.x509.X509Cert')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.fingerprint.domain_name.service.x509.x509_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=206
  _MESSAGE._serialized_end=331
# @@protoc_insertion_point(module_scope)



"""Ostorlab cli package."""

from ostorlab.cli import scan
from ostorlab.cli import auth
from ostorlab.cli import agent
from ostorlab.cli import vulnz
from ostorlab.cli import agentgroup
from ostorlab.cli import ci_scan
from ostorlab.cli import scanner

__all__ = ["scan", "auth", "agent", "vulnz", "agentgroup", "ci_scan", "scanner"]



"""Module for the command run inside the group scan.
This module takes care of preparing the selected runtime and the lists of provided agents, before starting a scan.
Example of usage:
    - ostorlab scan run --agent=agent1 --agent=agent2 --title=test_scan [asset] [options]."""

import io
import logging

import httpx
from typing import List

import click
from ruamel.yaml import error

from ostorlab import exceptions
from ostorlab.agent.schema import validator
from ostorlab.cli import console as cli_console
from ostorlab.cli import install_agent
from ostorlab.cli.scan import scan
from ostorlab.cli import types
from ostorlab.cli import agent_fetcher
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.utils import defintions as utils_definitions

console = cli_console.Console()

logger = logging.getLogger(__name__)


@scan.group(invoke_without_command=True)
@click.option(
    "--agent",
    multiple=True,
    help="List of agents keys. to use in the scan.",
    required=False,
)
@click.option(
    "--arg",
    multiple=True,
    help="""Add an argument to an agent. The argument should be in the format: <name>:<value>.
     Example: --arg fast_mode:true
    """,
    type=types.AgentArgType(),
    required=False,
)
@click.option("--title", "-t", help="Scan title.")
@click.option(
    "--agent-group-definition",
    "-g",
    type=click.File("r"),
    help="Path to agents group definition file (yaml).",
    required=False,
)
@click.option(
    "--assets",
    "-a",
    type=click.File("r"),
    help="Path to target list definition file (yaml).",
    required=False,
)
@click.option(
    "--install", "-i", help="Install missing agents.", is_flag=True, required=False
)
@click.option(
    "--follow",
    help="Follow logs of provided list of agents and services.",
    multiple=True,
    default=[],
)
@click.option(
    "--no-asset",
    help="Start the environment without injecting assets",
    is_flag=True,
    required=False,
)
@click.pass_context
def run(
    ctx: click.core.Context,
    agent: List[str],
    arg: list[types.AgentArg],
    agent_group_definition: io.FileIO,
    assets: io.FileIO,
    title: str,
    install: bool,
    follow: List[str],
    no_asset: bool,
) -> None:
    """Start a new scan on your assets.\n
    Example:\n
        - ostorlab scan run --agent=agent/ostorlab/nmap --agent=agent/google/tsunami --title=test_scan ip 8.8.8.8
    """
    if no_asset is True and ctx.invoked_subcommand is not None:
        console.error(
            f"Sub-command {ctx.invoked_subcommand} specified with --no-asset flag."
        )
        raise click.exceptions.Exit(2)
    if no_asset is False and ctx.invoked_subcommand is None and assets is None:
        console.error("Error: Missing command.")
        click.echo(ctx.get_help())
        raise click.exceptions.Exit(2)

    if agent:
        agents_settings: List[definitions.AgentSettings] = []
        for agent_key in agent:
            agents_settings.append(definitions.AgentSettings(key=agent_key))

        agent_group = definitions.AgentGroupDefinition(agents=agents_settings)
    elif agent_group_definition:
        try:
            agent_group = definitions.AgentGroupDefinition.from_yaml(
                agent_group_definition
            )
        except validator.ValidationError as e:
            console.error("Invalid agent group definition.")
            console.print(f"{e}")
            raise click.exceptions.Exit(2)
        except error.YAMLError as e:
            console.error("Agent group definition YAML parse error:")
            console.print(f"{e}")
            raise click.exceptions.Exit(2)
    else:
        raise click.ClickException("Missing agent list or agent group definition.")

    asset_group = None
    if assets is not None:
        try:
            asset_group = definitions.AssetsDefinition.from_yaml(assets)
        except validator.ValidationError as e:
            console.error(f"{e}")
            raise click.ClickException("Invalid asset Group Definition.") from e
    runtime_instance: runtime.Runtime = ctx.obj["runtime"]
    # set list of log follow.
    runtime_instance.follow = follow
    try:
        can_run_scan = runtime_instance.can_run(agent_group_definition=agent_group)
    except exceptions.OstorlabError as e:
        console.error(f"{e}")
        raise click.ClickException("Runtime encountered an error to run scan") from e

    if can_run_scan is True:
        ctx.obj["agent_group_definition"] = agent_group
        ctx.obj["title"] = title
        if install is True:
            try:
                # Trigger both the runtime installation routine and install all the provided agents.
                runtime_instance.install()
                for ag in agent_group.agents:
                    try:
                        install_agent.install(ag.key, ag.version)
                    except agent_fetcher.AgentDetailsNotFound:
                        console.warning(f"agent {ag.key} not found on the store")
            except httpx.HTTPError as e:
                raise click.ClickException(f"Could not install the agents: {e}")
        if arg is not None and len(arg) > 0:
            agent_group.agents = _add_cli_args_to_agent_settings(
                agent_group.agents, cli_args=arg
            )
        if ctx.invoked_subcommand is None:
            runtime_instance.scan(
                title=ctx.obj["title"],
                agent_group_definition=ctx.obj["agent_group_definition"],
                assets=asset_group if asset_group is None else asset_group.targets,
            )
    else:
        raise click.ClickException(
            "The runtime does not support the provided agent list or group definition."
        )


def _add_cli_args_to_agent_settings(
    agents_settings: list[definitions.AgentSettings], cli_args: list[types.AgentArg]
) -> list[definitions.AgentSettings]:
    """
    Adds CLI arguments to the agent settings if they are supported by the agent.

    Args:
        agents_settings : A list of agent settings.
        cli_args : A list of CLI Agent arguments.

    Returns:
         The updated list of agent settings.

    Raises:
        click.exceptions.Exit: If agent details are not found.
    """
    for agent_setting in agents_settings:
        try:
            agent_definition = agent_fetcher.get_definition(agent_setting.key)
        except agent_fetcher.AgentDetailsNotFound:
            console.error(f"Agent {agent_setting.key} not found.")
            raise click.exceptions.Exit(2)

        for cli_argument in cli_args:
            for arg in agent_definition.args:
                if arg.get("name") == cli_argument.name:
                    try:
                        agent_setting.args.append(
                            utils_definitions.Arg.build(
                                name=cli_argument.name,
                                type=arg.get("type", "string"),
                                value=cli_argument.value,
                            )
                        )
                        break
                    except ValueError as e:
                        console.warning(
                            f"Could not set argument {cli_argument.name} to {cli_argument.value} because of "
                            f"the following error: {e}"
                        )
    return agents_settings



"""This module contains custom types for the CLI commands."""

import dataclasses
from typing import Optional

import click


@dataclasses.dataclass
class AgentArg:
    """Dataclass representing an agent argument passed from cli."""

    name: str
    value: str


class AgentArgType(click.ParamType):
    """Custom Click type for parsing agent arguments in the format 'arg_name:arg_value'."""

    name = "arg_name:arg_value"

    def convert(
        self,
        arg_value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> AgentArg:
        """Override convert method from ParamType class to parse command line arguments of agents in the format
        'arg_name:arg_value' and convert them into AgentArg objects.

        Args:
            arg_value: The argument value in the format `arg_name:arg_value`.
            param: The parameter using this type for conversion. Defaults to None.
            ctx: The current click context. Defaults to None.

        Returns:
            AgentArg: An AgentArg object representing the parsed argument.

        Raises:
            click.BadParameter: If the argument value cannot be parsed into the expected format.
        """
        try:
            arg_name, arg_value = arg_value.split(":")
            return AgentArg(name=arg_name, value=arg_value)
        except ValueError:
            self.fail(
                message=f"Invalid argument {arg_value}. The expected format is name:value. Example: --arg "
                f"fast_mode:false",
                param=param,
                ctx=ctx,
            )
